% Sorting slrn's groups                  % -*- mode: slang; mode: fold -*-
%
% Copyright  2001-2003 J.B. Nicholson-Owens
% Author: J.B. Nicholson-Owens <jbn@forestfield.org>
% Title: Group Sort
% Version: 2.20
% Last tested with: slrn 0.9.7.4pl2 & S-Lang 1.4.9
%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program; if not, write to the Free Software
% Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  You may
% also download a copy of the applicable license from
% <URL:http://forestfield.org/gpl.txt>.

#iffalse % Documentation %{{{

These macros allow you to sort slrn's group list quickly and easily.  Bind
these common sorting algorithms to keys in group mode and sort the groups in
a variety of ways.

Much of the functionality of this macro set has become obsoleted by
set_group_order() and get_group_order().  Unless there are some particularly
interesting sorting techniques this macro set won't receive much attention.
The only real feature this macro set has now is to sort by last entered.

To install this macro set:

(1) Save a copy of this file (or article if you're getting this from Usenet)
    as "group_sort.sl".  Place this file somewhere so slrn can find it. Your
    home directory will do if you don't have a better place for it.

(2) Append the following lines to your slrn.rc ($HOME/.slrnrc on Unix) to
    make slrn load this file and set up some keys to access the functions:

       interpret "group_sort.sl"
       setkey group Group_Sort->sort_alphabetically   "^Ka"
       setkey group Group_Sort->sort_reverse          "^Kr"
       setkey group Group_Sort->toggle_last_entered   "^Ke"
       setkey group Group_Sort->sort_by_file          "^K%"
       setkey article Group_Sort->toggle_last_entered "^Ke"

    You can change the keymapping to whatever you like, these are only
    examples.

(3) Quit slrn and restart slrn.  The keys you mapped to Group Sort macros
    will take effect.

To remove this macro set (uninstall):

(1) Comment out or delete the interpret and setkey lines from your slrn.rc:

        % interpret "group_sort.sl"
        % setkey article "Group_Sort->sort_alphabetically "^Ka"
        % setkey article "Group_Sort->sort_reverse        "^Kr"
             .
             .

    Programs using Group_Sort functions will have to be altered so they are
    no longer dependant on Group_Sort or you may continue loading Group Sort
    but not assign any keys to its functions.

(2) Quit slrn and restart slrn.  The keys you mapped to Group Sort macros
    will revert back to whatever they were bound to before you started using
    Group Sort.

------------------------------------------------------------------------------

Function explanations:

    Void Group_Sort->toggle_last_entered():

        Use this function to turn on and off sorting by the order in which
        you read newsgroups.  This toggle defaults to off.

        Every time you enter a newsgroup the previous newsgroup you read is
        moved to the bottom of the newsgroup list.  As you read news this
        eventually produces a newsgroup list that is sorted by the order in
        which you read the newsgroups.  Sorting this way makes it easy to
        learn the order in which you prefer to read newsgroups.

    Void Group_Sort->sort_alphabetically

        Immediately sort the group list alphabetically.

    Void Group_Sort->sort_reverse

        Immediately reverse the order of the groups in the group list.

    Void Group_Sort->sort_by_file (String_Type path)
    Void Group_Sort->sort_by_file (File_Type id)
    Void Group_Sort->sort_by_file ()

         This function allows you to specify a file containing simple
         sorting sorting instructions.  The file can be specified by
         pathname (the first syntax), an already opened File_Type id (the
         second syntax), or (if you specify no argument) interactively
         by prompting for the sorting instruction filename.

         The syntax of the sorting instructions is as follows:

             % comment
             # comment
             regular-expression
             regular-expression last
             regular-expression before regular-expression
             regular-expression after regular-expression
             regular-expression first

         Comments can not appear on the same line as any other instruction:

              % THIS IS A COMMENT.
              regular-expression % THIS IS NOT A COMMENT.
              regular-expression first % THIS IS NOT A COMMENT.
              regular-expression last # THIS IS NOT A COMMENT.
              regular-expression before regular-expression % THIS IS NOT A COMMENT.
              regular-expression after regular-expression # THIS IS NOT A COMMENT.

         A regular expression alone is the same as appending the word "last"
         after that regular expression.  The keyword "last" means to sort
         matching groups last in the group list.  A series of regular
         expressions one before the other will sort the matching newsgroups
         in the order listed by the regular expressions.

         "before" and "after" only move groups that must be moved to satisfy
         the sort requirements.  "before" moves groups matching the left
         regular expression to the position before the topmost group
         matching the right regular expression.  "after" moves groups
         matching the left regular expression to the position after the
         lowest group matching the right regular expression.

         For example, if the group list were the letters of the alphabet in
         alphabetical order, "[a-c] before [q-t]" would move the groups a,
         b, and c before the group q.  Thus the group sequence would read
         {defghijklmnopabcqrstuvwxyz}.

         All of the commands in the file are interpreted immediately upon
         reading.  No state will be maintained across lines in the list of
         instructions.  This approach makes it simple to resolve circular
         dependency issues (like asking for "a before b" and "b before a").
         But it means a regular expression construct becomes a no-op.
         Because there is no state preserved between sorting instructions,
         you cannot refer to prior sorting decisions.  Consider the
         following:

               ^\(alpha-[a-z]+\)
               \$1\.discuss

         This will fail to do what the user probably expects because \$1
         means nothing.  The user was trying to capture the "alpha-"
         expression from the first sorting instruction and pass it to the
         second.  Perhaps this could be improved in the future to allow for
         passing state via regular expression registers.

Thanks for trying Group Sort.

2.20: * Added sorting by posted-to order.  The UI is not quite done yet.
        (2003-05-13)
      * Added interactive query for sorting by file to match docs.  Oops.
        (2003-07-05)

2.12: * Clarified and corrected the documentation.  Made bullet lists for
        the changelog and kept track of date of updates. (2003-04-13)
      * Simplified sorting by last-newsgroup-entered.  Now it works with one
        hook and has no macro wrappers for intrinsic functions.  Yay.
        (2003-04-13)

2.11: Fixed a bug in last().

2.1:  Added sort_by_file(), first(), last(), before(), and after() to sort by
      regular expression matching and easy to use commands.

2.0:  * Updated to use get_group_order() and set_group_order() which are much
        faster.
      * No longer requires any other macro set.  This macro set is stand-alone.

1.0:  Released. (2001)

#endif %}}}

implements ("Group_Sort");

% State and functions needed to implement sorting by last-entered-newsgroup.
private variable previous_entered_group = NULL;
static define article_mode_hook ()               %{{{
{
%   % If the switch is off, we do nothing.
%   !if (is_sorting_by_last_entered) return;

   if (current_newsgroup () != previous_entered_group)
     {
        % Move the previous_entered_group to another location in the newsgroup list.
        % Here we'll move the last entered newsgroup to the bottom of the
        % newsgroup list so the newsgroup list is eventually sorted by reading
        % order.
        variable group_order = get_group_order ();
        set_group_order (group_order[where (group_order != previous_entered_group)]);
        previous_entered_group = current_newsgroup ();
     }
}
%!if (register_hook ("article_mode_hook", "Group_Sort->article_mode_hook"))
%  message ("Group_Sort: I couldn\'t install what I needed to make sorting by last-read work.\nThis is a not a fatal error, so I'll keep going.");
%}}}
private variable is_sorting_by_last_entered = 0;
static define toggle_last_entered ()             %{{{
{
   if (is_sorting_by_last_entered)
     {
        % 0: Hook does not exist or function is not assigned to it.
        % 1: Function successfully unregistered.
        switch (unregister_hook ("article_mode_hook", "Group_Sort->article_mode_hook"))
          { case 0 : error ("Can\'t unregister hook.  article_mode_hook() does not exist in your build of slrn."); }
          { case 1 : message ("Not sorting by last-entered-newsgroup."); }
     }
   else
     {
        % 0: Hook does not exist or may not be defined multiple times.
        % 1: Function successfully registered.
        % 2: Given function was already registered for this hook.
        % 3: Undefined function successfully registered.
        switch (register_hook ("article_mode_hook", "Group_Sort->article_mode_hook"))
          { case 0 : error ("Can\'t register hook.  article_mode_hook() does not exist in your build of slrn."); }
          { orelse { case 1 } { case 2 } : message ("Sorting by last-entered-newsgroup."); }
          { case 3 : error ("Group_Sort->article_mode_hook was not seen by Global->register_hook()."); }
     }
   % Re-initialize the pointer to the last-entered-newsgroup.
   previous_entered_group = NULL;
   % Toggle the pointer to indicate whether we're sorting by
   % last-entered-newsgroup or not.
   is_sorting_by_last_entered = not (is_sorting_by_last_entered);
}
%}}}

% State and functions needed to implement sorting by last-posted-to-newsgroup.
static define post_filter_hook (file)
{
   % [[In the future, slrn will gain the ability to do this without requiring
   % manual input (pressing "f" to run the filter hook).  At that point this
   % function should switch to use that hook.]]

   % This function parses the file, extracts the list of newsgroups the
   % article was posted to (or the list of newsgroups where followups are
   % directed, favoring the followup list), and shuffles each of those
   % newsgroups (in that order) to the top of the newsgroup list.  This will
   % make it easier for the cursor to find the groups likely to contain
   % followups to those posts when a new slrn session begins.

   variable fp = fopen (file);
   if (fp == NULL) return;

   variable line = "", newsgroups = String_Type[0];
   while (-1 != fgets (&line, fp))
     {
        if (string_match ("^Followup-To: ", line, 1))
          {
             % This is the preferred set of groups to move to the top.
             % We'll break after this processing is done.
             (line, ) = strreplace (line, ", ", ",", strlen (line));
             % "13" below is the length of the string "Followup-To: ".
             newsgroups = strchopr (substr (line, 13, -1), ',', 0);

             break;
          }
        else if (string_match ("^Newsgroups: ", line, 1))
          {
             % This is the non-preferred set of groups to move to the top.
             % We'll break after this processing is done.
             (line, ) = strreplace (line, ", ", ",", strlen (line));
             % "13" below is the length of the string "Newsgroups: ".
             newsgroups = strchopr (substr (line, 13, -1), ',', 0);

             break;
          }
     }

   if (-1 == fclose (fp)) verror ("Could not close filehandle to %s", file);

   % Put each newsgroup at the top of the slrn newsgroup list.
   foreach (newsgroups) set_group_order ( () );
}

% Stand-alone sorting routines.
static define sort_alphabetically ()             %{{{
{
   % Sort the groups alphabetically.

   message_now ("Working...");
   variable group_order = get_group_order ();
   set_group_order (group_order [array_sort (group_order, &strcmp)]);
   message_now ("Working...done");
}

%}}}
static define sort_reverse ()                    %{{{
{
   % Reverse the order of the groups.

   message_now ("Working...");
   set_group_order (get_group_order ()[[-1:0:-1]]);
   message_now ("Working...done");
}

%}}}

% Routines typically used by Group_Sort->sort_by_file() but can be
% used stand-alone.
static define first (regex1)                     %{{{
{
   % Move the matching groups to the top of the list.

   variable group_order = get_group_order ();
   set_group_order (group_order[where (array_map (Int_Type, &string_match, group_order, regex1, 1))]);
}

%}}}
static define last (regex1)                      %{{{
{
   % Move the matching groups to the end of the list.

   variable
     % The list of newsgroups in their listed order.
     group_order = get_group_order (),
   
     % The newsgroups that match regex1.
     matching_groups = array_map (Int_Type, &string_match, group_order, regex1, 1),
   
     % The newsgroups that do NOT match regex1.
     nonmatching_groups = group_order[where (matching_groups == 0)];

   % set_group_order() lets us quickly shift groups to the top of the list.
   % We can exploit this to shift the groups we want to the bottom by giving
   % set_group_order() a list of groups that does not include the groups we
   % want to move to the bottom.  See the set_group_order() documentation for
   % more.
   set_group_order (nonmatching_groups);
}

%}}}
static define before (regex1, regex2)            %{{{
{
   % Move groups matching regex1 above the topmost group matching regex2.

   variable
     % The list of newsgroups.
     group_order = get_group_order ();

   variable
     % The list of newsgroups that match regex1 and regex2.
     r1_groups = group_order[where (array_map (Int_Type, &string_match, group_order, regex1, 1))],
     r2_groups = group_order[where (array_map (Int_Type, &string_match, group_order, regex2, 1))];

   % If there are no groups to move or no groups to compare against, there
   % is no work to do.
   !if (length (r1_groups)) return;
   !if (length (r2_groups)) return;

   % The index (in group_order) of the first newsgroup that matches regex2.
   variable r2_index = where (group_order == r2_groups[0])[0];

   % If r2_index == 0, we'll move the regex1 set of groups up to the top and stop.
   !if (r2_index) return first (regex1);

   % Otherwise we'll make a new string array consisting of (in order):
   %   The group(s) before the first r2_index group,
   %   The groups matching regex1,
   %   The groups matching regex2.
   variable sorted_groups = [group_order[[0 : r2_index - 1]], r1_groups, r2_groups];
   reshape (sorted_groups, [length (sorted_groups)]);
   set_group_order (sorted_groups);
}

%}}}
static define after (regex1, regex2)             %{{{
{
   % Move groups matching regex1 to after the bottom-most group matching
   % regex2.

   variable
     group_order = get_group_order (),
     r1_groups = group_order[where (array_map (Int_Type, &string_match, group_order, regex1, 1))],
     r2_groups = group_order[where (array_map (Int_Type, &string_match, group_order, regex2, 1))];

   !if (length (r1_groups)) return;
   !if (length (r2_groups)) return;

   % The index (in group_order) of the first newsgroup that matches regex2.
   variable
     r2_index = where (group_order == r2_groups[-1])[0],
     sorted_groups = [group_order[[0:r2_index]], r1_groups];

   set_group_order (sorted_groups);
}

%}}}
static variable previous_sort_by_file_instruction_file = make_home_filename (".slrn/newsgroup-sorting-instructions");
static define sort_by_file ()                    %{{{
{
   variable file = "";

   !if (_NARGS)
     {
        file = read_mini_filename ("Instruction file: ", "", previous_sort_by_file_instruction_file);
        previous_sort_by_file_instruction_file = file;
     }
   else file = ();

   message_now ("Working...");

   variable
     fp = fopen (file, "r"),
     line = "", words = String_Type[0],
     regex1 = "", command = "", regex2 = "";

   if (fp == NULL) verror ("Group_Sort: Could not open %s: %s", file, errno_string (errno));
   foreach (fp) using ("wsline")
     {
        line = strtrim ( () );

        if (string_match (line, "^[#%]", 1)) continue;
        words = strtok (line);

        switch (length (words))
          { case (1) : % Regex only -- move the matching groups to the end of the list.
             regex1 = words[0];
             last (regex1);
          }
          { case (2) : % Handle simple commands ("first", "last")
             regex1 = words[0];
             command = words[1];

             switch (command)
               { case ("first")  : first (regex1); }
               { case ("last")   : last (regex1); }
          }
          { case (3) : % Handle complex commands ("before", "after")
             regex1 = words[0];
             command = words[1];
             regex2 = words[2];

             switch (command)
               { case ("before") : before (regex1, regex2); }
               { case ("after")  : after (regex1, regex2); }
          }
     }

   message_now ("Working...done");
}

%}}}

static define version () { return (2.12); }
